module BDM (bDM) where

import qualified Data.Matrix as M
import qualified Data.Vector as V
import Data.List

--
import UD

data Quadrant = UL | LL | UR | LR deriving(Show, Eq)
quadL = [UL, LL, UR, LR]

-- BDM

-- Function that computes the BDM of a matrix.
bDM m d = minimum $ map f (quadLn d)
    where
    f qs = bDM' $ componentsM m d qs

bDM' (x:xs) = (logBase 2 (fromIntegral n)) + (uD m) + bDM' xs
            where
            (m, n) = x
bDM' [] = 0

-- A list with all combinations of quadrants of size
-- k.
quadLn 0 = [[]] 
quadLn k = [ q:l | q <- quadL, l <- quadLn (k - 1)] 

-- Debug&test matrix.
prueba2 = M.fromList (3+1) (3+2) [(1::Int)..] --(repeat (1::Int))
a12 = M.fromLists [[1::Int, 1, 1, 0, 1, 1, 1, 0], [0, 0, 1, 0, 0, 0, 0, 1], [0, 1, 1, 1, 1, 
  0, 1, 1], [1, 0, 0, 0, 1, 0, 0, 0], [1, 1, 0, 1, 1, 0, 1, 1], [0, 0,
   0, 1, 1, 0, 0, 0], [0, 1, 0, 1, 1, 0, 1, 0], [0, 1, 1, 0, 0, 1, 1, 
  0]]


-- Partition Algorithm.
part m d (q:qs) | d == 0 = []
           | (px == 0) || (py == 0) = part m (d-1) qs
           |otherwise = l ++ (resL q) ++ (resR q) ++ (resLR q)
   where
    maxP UL = M.submatrix 1 (d*py) 1 (d*px) m
    maxP LL = M.submatrix (qy + 1) y 1 (d*px) m
    maxP UR = M.submatrix 1 (d*py) (qx+1) (qx + d*px) m
    maxP LR = M.submatrix (qy + 1) (qy + d*py) (qx+1) (qx + d*px) m
    --
    resL UL = f (y + 1 - qy) y 1 (d*px)
    resL LL = f 1 qy 1 (d*px)
    resL UR = f 1 (d*py) 1 qx
    resL LR = f (qy + 1) y 1 qx
    --
    resR UL = f 1 (d*py) (d*px+1) x
    resR LL = f (qy+1) y (d*px+1) x
    resR UR = f (y + 1 - qy) y (qx + 1) x
    resR LR = f 1 qy (qx + 1) x
    --
    resLR UL = f (y + 1 - qy) y (x + 1 - qx) x
    resLR LL = f 1 qy (x + 1 - qx) x
    resLR UR = f (y + 1 - qy) y 1 qx
    resLR LR = f 1 qy 1 qx
    --
    f y1 y2 x1 x2 | (y1 > y2) || (x1 > x2) = []
      | otherwise = part (M.submatrix y1 y2 x1 x2 m) (d-1) qs
    -- List of partitions at the principal component.
    l = [ M.submatrix (1+y*d) ((y+1)*d) (1+x*d) ((x+1)*d) $ maxP q 
            | x <- [0..(px-1)], y <- [0..(py-1)]]
    x = M.ncols m
    y = M.nrows m
    (px, qx) = quotRem x d
    (py, qy) = quotRem y d

-- Function that compares matrices.
-- (Need a find a more efficient function).
-- (Use a custom Data.Matrix to extract a vector).
compareM m1 m2 | lm1 > lm2 = GT
               | lm2 < lm1 = LT
               | otherwise = EQ
    where
    lm1 = M.toList m1
    lm2 = M.toList m2
    
componentsM m d ps = components list
    where
    list = sortBy compareM (part m d ps)
    
    
-- Given an ORDERED list, find its components 
-- and their multiplicity.
components [] = []
components l = components' xs x 1 0
    where
    (x:xs) = l
    
components' (x:xs) y  c  0 | x == y = components' xs x (c+1) 0
                           | otherwise = (y, c) : components' xs x (1) 1
                           
components' (x:xs) y  c  1 | x == y = components' xs x (c+1) 0
                           | otherwise = (y, c) : components' xs x (1) 1

components' [] x c _ = [(x, c)]